/* Author: Ruud Harmsen.
   Copyright &copy; 2011, all rights reserved.
*/
/* Input (stdin): a list of paths to files, e.g. HTML files.
   Output (stdout): words in the input, one word per line,
      all converted to lowercase.

   A word is a sequence of at least one alphabetic character,
   followed by at least one alpha or digit. (So there's a minimum
   length of 2). Single quote ' and hyphen are also included.
   `deroff -w' does this too, but it doesn't support ISO-8859-1
   accented characters. Moreover, it isn't available in all
   Unix versions, e.g. FreeBSD doesn't have `deroff -w'.

   Enhancement 11 December 2013:
   To support some new Esperanto pages, I added support for
   entities (e.g. &ccirc;apitro for cxapitro or chapitro).
   It works, i.e., words containing such entities now appear
   in the index list at https://rudhar.com/index/index.htm .
   See also https://rudhar.com/index/cron/wordsep.htm .

   The sorting doesn't handle that entities right, e.g.
   &ccirc;apitro is sorted as if it really begins with &cc.

   Also, the coupling with my own search interface (s.c) to
   Swish (search index) doesn't properly the Esperanto words.

   Correction 22 December 2013:
   Things like '&amp' (without the closing semicolon) were
   interpreted as ampp, not amp, so the last letter was repeated.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

static int ExtractWords (FILE *fpi, FILE *fpo);
static int RemoveSomeEntities (char *wordbuf);
static int RemoveThisEntity (char *AfterAmpersand, char *OnSemicolon);

int main (int argc, char **argv)
{
   char filepath[129];
   FILE *fp;

   while (fgets(filepath, sizeof filepath, stdin) == filepath)
   {
      char *p;

      if (p = strrchr(filepath, '\n'))
         *p ='\0';
      if ((fp = fopen(filepath, "rt")) == NULL)
         continue;

      ExtractWords(fp, stdout);

      fclose(fp);
   }

   return 0;
}

static int ExtractWords (FILE *fpi, FILE *fpo)
{
   int inword = 0;
   int incmmd = 0;
   int inentity = 0;
   int c;
   unsigned char wordbuf[81];
   unsigned char *w = wordbuf;

   while (c = getc(fpi), c != EOF)
   {
      if (c == '<')
         incmmd = 1;
      else if (c == '>')
         incmmd = 0;

      if (!incmmd && !inword && (isalpha(c) || (c >= 0xc0 && c <= 0xff) || c == '&'))
      {
         inword = 1;
         w = wordbuf;
      }
      else if (inword &&
         !isalpha(c) && !isdigit(c) && c != '\'' &&
         c != '-' && !(c >= 0xc0 && c <= 0xff) && c != '&'
         && !(c == ';' && inentity) && !(c == '#' && inentity))
      {
         if (w > wordbuf + 1)
         {
            *w = '\0';

            /* Make the whole word lowercase */
            for (w = wordbuf; *w; w++)
            {
               /* tolower and toupper do work for ASCII, but not
                  for ISO-8859-1 accented letters. Flipping the third
                  bit from the left does.

                  Exception: the German sharp s is in the range for
                  uppercase and should not change, or it would
                  become a dotted y.
               */
               if (!(*w & 0x20) && *w != 0xDF)
                  *w = *w ^ 0x20;
            }
            RemoveSomeEntities(wordbuf);
            fprintf(fpo, "%s\n", wordbuf);
         }
         inword = 0;
         w = wordbuf;
      }

      if (inword)
      {
         if (w - wordbuf < sizeof wordbuf - 1)
            *w++ = c;

         if (c == '&')
            inentity = 1;
         else if (inentity && c == ';')
            inentity = 0;
      }
   }
   return 0;
}

static int RemoveSomeEntities (char *wordbuf)
{
   char *p, *m;

   /* Remove entities, like single or double curly quotes, from the start
      of the word buffer. */
   while (wordbuf[0] == '&')
   {
      m = wordbuf + strlen(wordbuf);
      for (p = wordbuf + 1; p < m && *p != ';'; p++)
         ;

      if (*p != ';')
      {
         /* Correction: shift also the terminating nullbyte */
         memmove(wordbuf, wordbuf + 1, m - wordbuf);
         break; /* Not a real entity, no terminating ';' found */
      }

      if (RemoveThisEntity(wordbuf + 1, p))
      {
         memmove(wordbuf, p + 1, m - p);
      }
      else
      {
         break;
      }
   }

   /* Remove entities, like single or double curly quotes, from the end
      of the word buffer. */
   m = wordbuf + strlen(wordbuf);
   while (m - wordbuf >= 4 && m[-1] == ';')
   {
      for (p = m - 2; p >= wordbuf && *p != '&'; p--)
         ;
      if (*p != '&')
      {
         m[-1] = '\0';
         break; /* Not a real entity, no starting & found */
      }
      if (RemoveThisEntity(p + 1, m - 1))
      {
         *p = '\0';
      }
      else
      {
         break;
      }
      m = wordbuf + strlen(wordbuf);
   }

   /* Remove any optional hyphens (suggested places to hyphenate) */
   while ((p = strstr(wordbuf, "&shy;")) != NULL)
   {
      memmove(p, p + strlen("&shy;"), m - p - strlen("&shy;") + 1);
   }

   return 0;
}

static char *ToStay[] =
{
   /* See http://etext.lib.virginia.edu/tei/isodia.html */
   "uml",   /* Umlaut and diaeresis (trema) */
   "acute",
   "grave",
   "circ",  /* Circumflex */
   "cedil", /* Cedille    */
   "tilde",
   "caron", /* For Slavic languages, also known as ha&ccaron;ek */
   "zlig",  /* &szlig; = scharfes S */
   "elig",  /* AE-ligature, ae, Danish, Norwegian, oe: French */
   "ring",  /* Ring above, for Skandinavian languages */
   "slash", /* Danish, Norwegian */
   "ogon",  /* Ogonek, for Polish */
   "dot",   /* Dot above, like for Polish z */
   "breve", /* Breve */
   "macr",  /* Macron */
   NULL
};
static int RemoveThisEntity (char *AfterAmpersand, char *OnSemicolon)
{
   int i;
   char tryent[26];
   int len;

   len = OnSemicolon - AfterAmpersand - 1;
   if (len + 1 > sizeof tryent)
      len = sizeof tryent - 1;
   strncpy(tryent, AfterAmpersand + 1, len);
   tryent[len] = '\0';

   /* Entities for special characters in several languages must stay
      because they are part of a word. */
   for (i = 0; ToStay[i]; i++)
   {
      if (strcmp(tryent, ToStay[i]) == 0)
         return 0;
   }

   /* Otherwise remove, things like &lsquo;, &rdquo;, &laquo;, &nbsp; */
   return 1;
}
